"""
General utilities for the MPESA functions
"""

from __future__ import print_function
from .exceptions import MpesaConfigurationException, IllegalPhoneNumberException, MpesaConnectionError, MpesaError
from django_daraja.models import AccessToken
import requests
from django.utils import timezone
from decouple import config, UndefinedValueError
import os
from requests import Response
import time
from django.conf import settings
import os
from django.conf import settings
import base64
from cryptography import x509
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15


class MpesaResponse(Response):
	response_description = ""
	error_code = None
	error_message = ''


def mpesa_response(r):
	"""
	Create MpesaResponse object from requests.Response object
	
	Arguments:
		r (requests.Response) -- The response to convert
	"""

	r.__class__ = MpesaResponse
	json_response = r.json()
	r.response_description = json_response.get('ResponseDescription', '')
	r.error_code = json_response.get('errorCode')
	r.error_message = json_response.get('errorMessage', '')
	return r


def mpesa_config(key):
	"""
	Get Mpesa configuration variable with the matching key
	
	Arguments:
		key (str) -- The configuration key

	Returns:
		str: Mpesa configuration variable with the matching key

	Raises:
		MpesaConfigurationException: Key not found
	"""

	value = getattr(settings, key, None)
	if value is None:
		try:
			value = config(key)
		except UndefinedValueError:
			# Check key in settings file
			raise MpesaConfigurationException('Mpesa environment not configured properly - ' + key + ' not found')

	return value


def api_base_url():
	"""
	Gets the base URL for making API calls

	Returns:
		The base URL depending on development environment (sandbox or production)

	Raises:
		MpesaConfigurationException: Environment not sandbox or production
	"""

	mpesa_environment = mpesa_config('MPESA_ENVIRONMENT')

	if mpesa_environment == 'sandbox':
		return 'https://sandbox.safaricom.co.ke/'
	elif mpesa_environment == 'production':
		return 'https://api.safaricom.co.ke/'
	else:
		raise MpesaConfigurationException('Mpesa environment not configured properly - MPESA_ENVIRONMENT should be sandbox or production')

def generate_access_token_request(consumer_key = None, consumer_secret = None):
	"""
	Make a call to OAuth API to generate access token
	
	Arguments:
		consumer_key (str) -- (Optional) The Consumer Key to use
		consumer_secret (str) -- (Optional) The Consumer Secret to use

	Returns:
		requests.Response: Response object with the response details

	Raises:
		MpesaConnectionError: Connection error
	"""

	url = api_base_url() + 'oauth/v1/generate?grant_type=client_credentials'
	consumer_key = consumer_key if consumer_key is not None else mpesa_config('MPESA_CONSUMER_KEY') 
	consumer_secret = consumer_secret if consumer_secret is not None else mpesa_config('MPESA_CONSUMER_SECRET')

	try:
		r = requests.get(url, auth=(consumer_key, consumer_secret))
	except requests.exceptions.ConnectionError:
		raise MpesaConnectionError('Connection failed')
	except Exception as ex:
		return ex.message
	
	return r

def generate_access_token():
	"""
	Parse generated OAuth access token, then updates database access token value

	Returns:
		AccessToken: The AccessToken object from the database

	Raises:
		MpesaError: Error generating access token
	"""

	r = generate_access_token_request()
	if r.status_code != 200:
		# Retry to generate access token
		r = generate_access_token_request()
		if r.status_code != 200:
			raise MpesaError('Unable to generate access token')

	token = r.json()['access_token']

	AccessToken.objects.all().delete()
	access_token = AccessToken.objects.create(token=token)

	return access_token

def mpesa_access_token():
	"""
	Generate access token if the current one has expired or if token is non-existent
	Otherwise return existing access token

	Returns:
		str: A valid access token
	"""

	access_token = AccessToken.objects.first()
	if access_token == None:
		# No access token found
		access_token = generate_access_token()
	else:
		delta = timezone.now() - access_token.created_at
		minutes = (delta.total_seconds()//60)%60
		print('minutes: ', minutes)
		if minutes > 50:
			# Access token expired
			access_token = generate_access_token()	
	
	return access_token.token

def format_phone_number(phone_number):
	"""
	Format phone number into the format 2547XXXXXXXX
	
	Arguments:
		phone_number (str) -- The phone number to format
	"""

	if len(phone_number) < 9:
		raise IllegalPhoneNumberException('Phone number too short')
	else:
		return '254' + phone_number[-9:]

def sleep(seconds, message=''):
	"""
	Sleep for the specified number of seconds
	
	Arguments:
		seconds (float) -- Number of seconds to sleep, can be a float
		message (str) -- (Optional) message to display
	"""

	print()
	print('===')
	print(message, end="")
	for i in range(seconds * 2):
		time.sleep(0.5)
		print('.', end="")
	print()
	print('===')
	print()

def encrypt_security_credential(credential):
	"""
	Generate an encrypted security credential from a plaintext value
	
	Arguments:
		credential (str) -- The plaintext credential display
	"""

	mpesa_environment = mpesa_config('MPESA_ENVIRONMENT')

	if mpesa_environment in ('sandbox', 'production'):
		certificate_name = mpesa_environment + '.cer'
	else:
		raise MpesaConfigurationException('Mpesa environment not configured properly - MPESA_ENVIRONMENT should be sandbox or production')

	certificate_path = os.path.join(settings.BASE_DIR, 'certs', certificate_name)
	return encrypt_rsa(certificate_path, credential)

def encrypt_rsa(certificate_path, input):
	message = input.encode('ascii')
	with open(certificate_path, "rb") as cert_file:
		cert = x509.load_pem_x509_certificate(cert_file.read())
		encrypted = cert.public_key().encrypt(message, PKCS1v15())
		output = base64.b64encode(encrypted).decode('ascii')

	return output


# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.urls import reverse
from django.test import TestCase

class MpesaCoreTestCase(TestCase):

	def test_index_view(self):
		'''
		Test home page content
		'''
		
		response = self.client.get(reverse("django_daraja_index"))
		self.assertEqual(response.content, b"Welcome to the home of daraja APIs")


# Generated by Django 2.1.2 on 2018-11-07 12:16

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='AccessToken',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('token', models.CharField(max_length=30)),
                ('created_at', models.DateField(auto_now_add=True)),
            ],
            options={
                'get_latest_by': 'created_at',
            },
        ),
    ]



from __future__ import absolute_import, unicode_literals
import os

test_runner = None
old_config = None

os.environ["DJANGO_SETTINGS_MODULE"] = "tests.settings"

import django
if hasattr(django, "setup"):
    django.setup()

def setup():
    global test_runner
    global old_config

    # If you want to support Django 1.5 and older, you need
    # this try-except block.
    try:
        from django.test.runner import DiscoverRunner
        test_runner = DiscoverRunner()
    except ImportError:
        from django.test.simple import DjangoTestSuiteRunner
        test_runner = DjangoTestSuiteRunner()

    test_runner.setup_test_environment()
    old_config = test_runner.setup_databases()

def teardown():
    test_runner.teardown_databases(old_config)
    test_runner.teardown_test_environment()


import requests
from requests.auth import HTTPDigestAuth
import base64
from datetime import datetime
import json
from .exceptions import MpesaInvalidParameterException, MpesaConnectionError
from .utils import encrypt_security_credential, mpesa_access_token, format_phone_number, api_base_url, mpesa_config, mpesa_response
from decouple import config

class MpesaClient:
	"""
	This is the core MPESA client. 

	The Mpesa Client will access all interactions with the MPESA Daraja API.
	"""

	auth_token = ''

	def __init__(self):
		"""
		The constructor for MpesaClient class
		"""

	def access_token(self):
		"""
		Generate an OAuth access token.

		Returns:
			bool: A string containg a valid OAuth access token
		"""
		
		return mpesa_access_token()

	def parse_stk_result(self, result):
		"""
		Parse the result of Lipa na MPESA Online Payment (STK Push)

		Returns:
			The result data as an array
		"""
		
		payload = json.loads(result)
		data = {}
		callback = payload['Body']['stkCallback']
		data['ResultCode'] = callback['ResultCode']
		data['ResultDesc'] = callback['ResultDesc']
		data['MerchantRequestID'] = callback['MerchantRequestID']
		data['CheckoutRequestID'] = callback['CheckoutRequestID']
		metadata = callback.get('CallbackMetadata')
		if metadata:
			metadata_items = metadata.get('Item')
			for item in metadata_items:
				data[item['Name']] = item.get('Value')
		
		return data

	def stk_push(self, phone_number, amount, account_reference, transaction_desc, callback_url):
		"""
		Attempt to send an STK prompt to customer phone

		Args:
			phone_number (str): -- The Mobile Number to receive the STK Pin Prompt.
			amount (int) -- This is the Amount transacted normaly a numeric value. Money that customer pays to the Shorcode. Only whole numbers are supported.
			account_reference (str) -- This is an Alpha-Numeric parameter that is defined by your system as an Identifier of the transaction for CustomerPayBillOnline transaction type. Along with the business name, this value is also displayed to the customer in the STK Pin Prompt message. Maximum of 12 characters.
			transaction_desc (str) -- This is any additional information/comment that can be sent along with the request from your system. Maximum of 13 Characters.
			call_back_url (str) -- This s a valid secure URL that is used to receive notifications from M-Pesa API. It is the endpoint to which the results will be sent by M-Pesa API.

		Returns:
			MpesaResponse: MpesaResponse object containing the details of the API response
		
		Raises:
			MpesaInvalidParameterException: Invalid parameter passed
			MpesaConnectionError: Connection error
		"""

		if str(account_reference).strip() == '':
			raise MpesaInvalidParameterException('Account reference cannot be blank')
		if str(transaction_desc).strip() == '':
			raise MpesaInvalidParameterException('Transaction description cannot be blank')
		if not isinstance(amount, int):
			raise MpesaInvalidParameterException('Amount must be an integer')


		phone_number = format_phone_number(phone_number)
		url = api_base_url() + 'mpesa/stkpush/v1/processrequest'
		passkey = mpesa_config('MPESA_PASSKEY')
		
		mpesa_environment = mpesa_config('MPESA_ENVIRONMENT')
		if mpesa_environment == 'sandbox':
			business_short_code = mpesa_config('MPESA_EXPRESS_SHORTCODE')
		else:
			business_short_code = mpesa_config('MPESA_SHORTCODE')

		timestamp = datetime.now().strftime('%Y%m%d%H%M%S')
		password = base64.b64encode((business_short_code + passkey + timestamp).encode('ascii')).decode('utf-8') 
		transaction_type = 'CustomerPayBillOnline'
		party_a = phone_number
		party_b = business_short_code

		data = {
			'BusinessShortCode': business_short_code,
			'Password': password,
			'Timestamp': timestamp,
			'TransactionType': transaction_type,
			'Amount': amount,
			'PartyA': party_a,
			'PartyB': party_b,
			'PhoneNumber': phone_number,
			'CallBackURL': callback_url,
			'AccountReference': account_reference,
			'TransactionDesc': transaction_desc
		}

		headers = {
			'Authorization': 'Bearer ' + mpesa_access_token(),
			'Content-type': 'application/json'
		}

		try:
			r = requests.post(url, json=data, headers=headers)
			response = mpesa_response(r)
			return response
		except requests.exceptions.ConnectionError:
			raise MpesaConnectionError('Connection failed')
		except Exception as ex:
			raise MpesaConnectionError(str(ex))

	def b2c_payment(self, phone_number, amount, transaction_desc, callback_url, occassion, command_id):
		"""
		Attempt to perform a business payment transaction

		Args:
			phone_number (str): -- The Mobile Number to receive the STK Pin Prompt.
			amount (int) -- This is the Amount transacted normaly a numeric value. Money that customer pays to the Shorcode. Only whole numbers are supported.
			transaction_desc (str) -- This is any additional information/comment that can be sent along with the request from your system. Maximum of 13 Characters.
			call_back_url (str) -- This s a valid secure URL that is used to receive notifications from M-Pesa API. It is the endpoint to which the results will be sent by M-Pesa API.
			occassion (str) -- Any additional information to be associated with the transaction.

		Returns:
			MpesaResponse: MpesaResponse object containing the details of the API response
		
		Raises:
			MpesaInvalidParameterException: Invalid parameter passed
			MpesaConnectionError: Connection error
		"""

		if str(transaction_desc).strip() == '':
			raise MpesaInvalidParameterException('Transaction description cannot be blank')
		if not isinstance(amount, int):
			raise MpesaInvalidParameterException('Amount must be an integer')

		phone_number = format_phone_number(phone_number)
		url = api_base_url() + 'mpesa/b2c/v1/paymentrequest'

		business_short_code = mpesa_config('MPESA_SHORTCODE')

		party_a = business_short_code
		party_b = phone_number
		initiator_username = mpesa_config('MPESA_INITIATOR_USERNAME')
		initiator_security_credential = encrypt_security_credential(mpesa_config('MPESA_INITIATOR_SECURITY_CREDENTIAL'))

		data = {
			'InitiatorName': initiator_username,
			'SecurityCredential': initiator_security_credential,
			'CommandID': command_id,
			'Amount': amount,
			'PartyA': party_a,
			'PartyB': party_b,
			'Remarks': transaction_desc,
			'QueueTimeOutURL': callback_url,
			'ResultURL': callback_url,
			'Occassion':  occassion
		}

		headers = {
			'Authorization': 'Bearer ' + mpesa_access_token(),
			'Content-type': 'application/json'
		}

		try:
			r = requests.post(url, json=data, headers=headers)
			response = mpesa_response(r)
			return response
		except requests.exceptions.ConnectionError:
			raise MpesaConnectionError('Connection failed')
		except Exception as ex:
			raise MpesaConnectionError(str(ex))

	def business_payment (self, phone_number, amount, transaction_desc, callback_url, occassion):
		command_id = 'BusinessPayment'
		return self.b2c_payment(phone_number, amount, transaction_desc, callback_url, occassion, command_id)

	def salary_payment (self, phone_number, amount, transaction_desc, callback_url, occassion):
		command_id = 'SalaryPayment'
		return self.b2c_payment(phone_number, amount, transaction_desc, callback_url, occassion, command_id)

	def promotion_payment (self, phone_number, amount, transaction_desc, callback_url, occassion):
		command_id = 'PromotionPayment'
		return self.b2c_payment(phone_number, amount, transaction_desc, callback_url, occassion, command_id)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from django_daraja.mpesa import utils

from django.http import HttpResponse, JsonResponse
from django.views.generic import View
from django_daraja.mpesa.core import MpesaClient
from decouple import config
from datetime import datetime

cl = MpesaClient()
stk_push_callback_url = 'https://darajambili.herokuapp.com/express-payment'
b2c_callback_url = 'https://darajambili.herokuapp.com/b2c/result'

def index(request):

	return HttpResponse('Welcome to the home of daraja APIs')

def oauth_success(request):
	r = cl.access_token()
	return JsonResponse(r, safe=False)

def stk_push_success(request):
	phone_number = config('LNM_PHONE_NUMBER')
	amount = 1
	account_reference = 'ABC001'
	transaction_desc = 'STK Push Description'
	callback_url = stk_push_callback_url
	r = cl.stk_push(phone_number, amount, account_reference, transaction_desc, callback_url)
	return JsonResponse(r.response_description, safe=False)

def business_payment_success(request):
	phone_number = config('B2C_PHONE_NUMBER')
	amount = 1
	transaction_desc = 'Business Payment Description'
	occassion = 'Test business payment occassion'
	callback_url = b2c_callback_url
	r = cl.business_payment(phone_number, amount, transaction_desc, callback_url, occassion)
	return JsonResponse(r.response_description, safe=False)

def salary_payment_success(request):
	phone_number = config('B2C_PHONE_NUMBER')
	amount = 1
	transaction_desc = 'Salary Payment Description'
	occassion = 'Test salary payment occassion'
	callback_url = b2c_callback_url
	r = cl.salary_payment(phone_number, amount, transaction_desc, callback_url, occassion)
	return JsonResponse(r.response_description, safe=False)

def promotion_payment_success(request):
	phone_number = config('B2C_PHONE_NUMBER')
	amount = 1
	transaction_desc = 'Promotion Payment Description'
	occassion = 'Test promotion payment occassion'
	callback_url = b2c_callback_url
	r = cl.promotion_payment(phone_number, amount, transaction_desc, callback_url, occassion)
	return JsonResponse(r.response_description, safe=False)



# -*- coding: utf-8 -*-
'''
Test Business Payment
'''

from django.test import TestCase
from django_daraja.mpesa.core import MpesaClient
from decouple import config
from django_daraja.mpesa.exceptions import MpesaInvalidParameterException
from django_daraja.mpesa.utils import sleep

class MpesaB2CPaymentTestCase(TestCase):

	cl = MpesaClient()
	callback_url = 'https://darajambili.herokuapp.com/express-payment'
	success_description = 'Accept the service request successfully.'

	def test_business_payment_success(self):
		'''
		Test successful business payment
		'''

		# Wait for a short while (to avoid SpikeArrest)
		sleep(20, 'Test business payment success')

		phone_number = config('B2C_PHONE_NUMBER')
		amount = 1
		transaction_desc = 'Description'
		occassion = 'Occassion'
		response = self.cl.business_payment(phone_number, amount, transaction_desc, self.callback_url, occassion)
		self.assertEqual(response.response_description, self.success_description)

	def test_salary_payment_success(self):
		'''
		Test successful salary payment
		'''
		
		# Wait for a short while (to avoid SpikeArrest)
		sleep(20, 'Test salary payment success')

		phone_number = config('B2C_PHONE_NUMBER')
		amount = 1
		transaction_desc = 'Description'
		occassion = 'Occassion'
		response = self.cl.salary_payment(phone_number, amount, transaction_desc, self.callback_url, occassion)
		self.assertEqual(response.response_description, self.success_description)

	def test_promotion_payment_success(self):
		'''
		Test successful promotion payment
		'''

		# Wait for a short while (to avoid SpikeArrest)
		sleep(20, 'Test promotion payment success')

		phone_number = config('B2C_PHONE_NUMBER')
		amount = 1
		transaction_desc = 'Description'
		occassion = 'Occassion'
		response = self.cl.promotion_payment(phone_number, amount, transaction_desc, self.callback_url, occassion)
		self.assertEqual(response.response_description, self.success_description)
				
	def test_business_payment_empty_description(self):
		'''
		Test that Business Payment with empty description raises MpesaInvalidParameterException
		'''

		with self.assertRaises(MpesaInvalidParameterException):
			phone_number = config('B2C_PHONE_NUMBER')
			amount = 1000000
			occassion = 'occassion'
			transaction_desc = ''
			self.cl.business_payment(phone_number, amount, transaction_desc, self.callback_url, occassion)

	def test_business_payment_invalid_amount(self):
		'''
		Test that Business Payment with invalid amount raises MpesaInvalidParameterException
		'''

		with self.assertRaises(MpesaInvalidParameterException):
			phone_number = config('B2C_PHONE_NUMBER')
			amount = 1.5
			occassion = 'occassion'
			transaction_desc = 'Description'
			self.cl.business_payment(phone_number, amount, transaction_desc, self.callback_url, occassion)



# -*- coding: utf-8 -*-
"""
Test the MPESA client
"""

from __future__ import unicode_literals
from django.urls import reverse
from django.test import TestCase
from django_daraja.mpesa.core import MpesaClient

class MpesaClientTestCase(TestCase):

	def test_client_init(self):
		'''
		Test initialization of MPESA Client
		'''
		
		cl = MpesaClient()
		self.assertEqual(cl.__class__.__name__, 'MpesaClient')




# Generated by Django 2.1.2 on 2018-11-08 09:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_daraja', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='accesstoken',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True),
        ),
    ]



"""
Exceptions raised during usage of the MPESa API
"""

class MpesaError(Exception):
	"""
	Raised for a general error regarding the library
	"""

class IllegalPhoneNumberException(Exception):
	"""
	Raised when phone number is in illegal format.
	"""
	pass


class MpesaConnectionError(Exception):
	"""
	Raised when connection has an error
	"""
	pass

class MpesaConfigurationException(Exception):
	"""
	Raised when Mpesa environment variables are not configured properly
	"""
	pass

class MpesaInvalidParameterException(Exception):
	"""
	Raised when an an invalid parameter is passed in a function call
	
	Extends:
		Exception: Exception
	"""
	pass


